# coding: utf-8
from __future__ import annotations

import json
from enum import Enum
from typing import Optional

from supervisely.app import DialogWindowError
from supervisely.task.progress import Progress


class PullPolicy(Enum):
    def __str__(self):
        return str(self.value)

    ALWAYS = "Always".lower()
    IF_AVAILABLE = "IfAvailable".lower()
    IF_NOT_PRESENT = "IfNotPresent".lower()
    NEVER = "Never".lower()


class PullStatus(Enum):
    START = "Pulling fs layer"
    DOWNLOAD = "Downloading"
    EXTRACT = "Extracting"
    COMPLETE_LOAD = "Download complete"
    COMPLETE_PULL = "Pull complete"
    OTHER = "Other (unknown)"

    def is_equal(self, status: str) -> bool:
        return status == self.value

    @classmethod
    def from_str(cls, status: Optional[str]) -> PullStatus:
        dct = {
            "Pulling fs layer": PullStatus.START,
            "Downloading": PullStatus.DOWNLOAD,
            "Extracting": PullStatus.EXTRACT,
            "Download complete": PullStatus.COMPLETE_LOAD,
            "Pull complete": PullStatus.COMPLETE_PULL,
        }
        return dct.get(status, PullStatus.OTHER)


def docker_pull_if_needed(docker_api, docker_image_name, policy, logger, progress=True):
    logger.info(
        "docker_pull_if_needed args",
        extra={
            "policy": policy,
            "type(policy)": type(policy),
            "policy == PullPolicy.ALWAYS": str(policy) == str(PullPolicy.ALWAYS),
            "policy == PullPolicy.NEVER": str(policy) == str(PullPolicy.NEVER),
            "policy == PullPolicy.IF_NOT_PRESENT": str(policy) == str(PullPolicy.IF_NOT_PRESENT),
            "policy == PullPolicy.IF_AVAILABLE": str(policy) == str(PullPolicy.IF_AVAILABLE),
        },
    )
    if str(policy) == str(PullPolicy.ALWAYS):
        if progress is False:
            _docker_pull(docker_api, docker_image_name, logger)
        else:
            _docker_pull_progress(docker_api, docker_image_name, logger)
    elif str(policy) == str(PullPolicy.NEVER):
        pass
    elif str(policy) == str(PullPolicy.IF_NOT_PRESENT):
        if not _docker_image_exists(docker_api, docker_image_name):
            if progress is False:
                _docker_pull(docker_api, docker_image_name, logger)
            else:
                _docker_pull_progress(docker_api, docker_image_name, logger)
    elif str(policy) == str(PullPolicy.IF_AVAILABLE):
        if progress is False:
            _docker_pull(docker_api, docker_image_name, logger, raise_exception=True)
        else:
            _docker_pull_progress(docker_api, docker_image_name, logger, raise_exception=True)
    else:
        raise RuntimeError(f"Unknown pull policy {str(policy)}")
    if not _docker_image_exists(docker_api, docker_image_name):
        raise DialogWindowError(
            title=f"Docker image {docker_image_name} not found. Agent's PULL_POLICY is {str(policy)}.",
            description=(
                "The initiation of the pulling process was either prevented due to the pull policy settings "
                "or it was halted mid-way because the host lacks sufficient disk space."
            ),
        )


def _docker_pull(docker_api, docker_image_name, logger, raise_exception=True):
    from docker.errors import DockerException

    logger.info("Docker image will be pulled", extra={"image_name": docker_image_name})
    progress_dummy = Progress("Pulling image...", 1, ext_logger=logger)
    progress_dummy.iter_done_report()
    try:
        pulled_img = docker_api.images.pull(docker_image_name)
        logger.info(
            "Docker image has been pulled",
            extra={"pulled": {"tags": pulled_img.tags, "id": pulled_img.id}},
        )
    except DockerException as e:
        if raise_exception is True:
            raise DockerException(
                "Unable to pull image: see actual error above. "
                "Please, run the task again or contact support team."
            )
        else:
            logger.warn("Pulling step is skipped. Unable to pull image: {!r}.".format(str(e)))


def _docker_pull_progress(docker_api, docker_image_name, logger, raise_exception=True):
    logger.info("Docker image will be pulled", extra={"image_name": docker_image_name})
    from docker.errors import DockerException

    try:
        layers_total_load = {}
        layers_current_load = {}
        layers_total_extract = {}
        layers_current_extract = {}
        started = set()
        loaded = set()
        pulled = set()

        progress_full = Progress("Preparing dockerimage", 1, ext_logger=logger)
        progres_ext = Progress("Extracting layers", 1, is_size=True, ext_logger=logger)
        progress_load = Progress("Downloading layers", 1, is_size=True, ext_logger=logger)

        for line in docker_api.api.pull(docker_image_name, stream=True, decode=True):
            status = PullStatus.from_str(line.get("status", None))
            layer_id = line.get("id", None)
            progress_details = line.get("progressDetail", {})
            need_report = True

            if status is PullStatus.START:
                started.add(layer_id)
                need_report = False
            elif status is PullStatus.DOWNLOAD:
                layers_total_load[layer_id] = progress_details.get("total", 1)
                layers_current_load[layer_id] = progress_details.get("current", 0)
                total_load = sum(layers_total_load.values())
                current_load = sum(layers_current_load.values())
                if total_load > progress_load.total:
                    progress_load.set(current_load, total_load)
                elif (current_load - progress_load.current) / total_load > 0.01:
                    progress_load.set(current_load, total_load)
                else:
                    need_report = False
            elif status is PullStatus.COMPLETE_LOAD:
                loaded.add(layer_id)
            elif status is PullStatus.EXTRACT:
                layers_total_extract[layer_id] = progress_details["total"]
                layers_current_extract[layer_id] = progress_details["current"]
                total_ext = sum(layers_total_extract.values())
                current_ext = sum(layers_current_extract.values())
                if total_ext > progres_ext.total:
                    progres_ext.set(current_ext, total_ext)
                elif (current_ext - progres_ext.current) / total_ext > 0.01:
                    progres_ext.set(current_ext, total_ext)
                else:
                    need_report = False
            elif status is PullStatus.COMPLETE_PULL:
                pulled.add(layer_id)

            if started != pulled:
                if need_report:
                    if started == loaded:
                        progres_ext.report_progress()
                    else:
                        progress_load.report_progress()
            elif len(pulled) > 0:
                progress_full.report_progress()

        progress_full.iter_done()
        progress_full.report_progress()
        logger.info("Docker image has been pulled", extra={"image_name": docker_image_name})
    except DockerException as e:
        if raise_exception is True:
            raise e
            # raise DockerException(
            #     "Unable to pull image: see actual error above. "
            #     "Please, run the task again or contact support team."
            # )
        else:
            logger.warn("Pulling step is skipped. Unable to pull image: {!r}.".format(repr(e)))


def _docker_image_exists(docker_api, docker_image_name):
    from docker.errors import ImageNotFound

    try:
        docker_img = docker_api.images.get(docker_image_name)
    except ImageNotFound:
        return False
    return True



from collections import defaultdict
from typing import Any, Dict, List, Optional, Tuple
from uuid import uuid4

import numpy as np

from supervisely.annotation.annotation import Annotation
from supervisely.annotation.json_geometries_map import GET_GEOMETRY_FROM_STR
from supervisely.annotation.label import Label
from supervisely.annotation.obj_class import ObjClass
from supervisely.annotation.tag import Tag
from supervisely.annotation.tag_meta import TagMeta, TagValueType
from supervisely.geometry.bitmap import Bitmap
from supervisely.geometry.geometry import Geometry
from supervisely.geometry.point_location import row_col_list_to_points
from supervisely.geometry.polygon import Polygon
from supervisely.geometry.rectangle import Rectangle
from supervisely.project.project_meta import ProjectMeta
from supervisely.sly_logger import logger

POSSIBLE_SHAPE_TYPES = ["polygonlabels", "rectanglelabels", "brushlabels"]
POSSIBLE_TAGS_TYPES = ["choices"]


class InputStream:
    def __init__(self, data):
        self.data = data
        self.i = 0

    def read(self, size):
        out = self.data[self.i : self.i + size]
        self.i += size
        return int(out, 2)


def access_bit(data, num):
    """from bytes array to bits by num position"""
    base = int(num // 8)
    shift = 7 - int(num % 8)
    return (data[base] & (1 << shift)) >> shift


def bytes2bit(data):
    """get bit string from bytes data"""
    return "".join([str(access_bit(data, i)) for i in range(len(data) * 8)])


def convert_rle_mask_to_sly(coords: List[int], h: int, w: int) -> Optional[Bitmap]:
    """Convert Label Studio RLE mask to Supervisely Bitmap."""

    rle_input = InputStream(bytes2bit(coords))

    num = rle_input.read(32)
    word_size = rle_input.read(5) + 1
    rle_sizes = [rle_input.read(4) + 1 for _ in range(4)]

    i = 0
    out = np.zeros(num, dtype=np.uint8)
    while i < num:
        x = rle_input.read(1)
        j = i + 1 + rle_input.read(rle_sizes[rle_input.read(2)])
        if x:
            val = rle_input.read(word_size)
            out[i:j] = val
            i = j
        else:
            while i < j:
                val = rle_input.read(word_size)
                out[i] = val
                i += 1

    mask = np.reshape(out, [h, w, 4])[:, :, 3].astype(np.bool)
    if mask.ndim == 3:
        mask = mask[:, :, 0]

    return Bitmap(mask)


def convert_polygon_to_sly(coords: List[List[int]], h: int, w: int) -> Polygon:
    """Convert Label Studio polygon coordinates to Supervisely Polygon object."""

    coords = [[int(row * h / 100), int(col * w / 100)] for (col, row) in coords]
    points = row_col_list_to_points(coords, do_round=True)
    return Polygon(points)


def convert_rectangle_to_sly(coords: List[int], h: int, w: int) -> Rectangle:
    """Convert Label Studio rectangle coordinates to Supervisely Rectangle object."""

    x, y, rect_w, rect_h = coords
    r, b = x + rect_w, y + rect_h
    l, t, r, b = int(x * w / 100), int(y * h / 100), int(r * w / 100), int(b * h / 100)
    return Rectangle(t, l, b, r)


convert_geom_func_map = {
    "polygonlabels": (convert_polygon_to_sly, ("points",)),
    "rectanglelabels": (convert_rectangle_to_sly, ("x", "y", "width", "height")),
    "brushlabels": (convert_rle_mask_to_sly, ("rle",)),
}


def convert_geometry_to_sly(
    shape_type: str, data: Dict[str, Any], h: int, w: int
) -> Optional[Geometry]:
    try:
        convert_func, keys = convert_geom_func_map.get(shape_type, (None, None))
        if convert_func is None:
            raise ValueError(f"Unsupported shape type: {shape_type}")
        coords = [data.get(key) for key in keys] if len(keys) > 1 else data.get(keys[0])
        geometry = convert_func(coords, h, w)
        return geometry
    except Exception as e:
        logger.warn(f"Failed to convert geometry: {shape_type}. Reason: {repr(e)}")
        return None


def generate_new_name_for_meta(
    meta: ProjectMeta,
    name: str,
    expected_type: type,
    i: Optional[int] = None,
    is_tag: bool = False,
) -> str:
    """Generate new name for tag or class in project meta."""

    new_name = f"{name}_{i}" if i else name
    item = meta.get_tag_meta(new_name) if is_tag else meta.get_obj_class(new_name)
    if item is None:
        return new_name
    if is_tag and item.value_type == expected_type:
        return new_name
    if not is_tag and item.geometry_type == expected_type:
        return new_name
    return generate_new_name_for_meta(meta, name, expected_type, i + 1 if i else 1, is_tag)


def _get_or_create_tag_or_cls(meta: ProjectMeta, name: str, geometry_name: Optional[str] = None):
    """Private function to get or create tag or class in project meta."""

    is_tag = geometry_name is None
    item_type = TagValueType.NONE if is_tag else GET_GEOMETRY_FROM_STR(geometry_name)
    item = meta.get_tag_meta(name) if is_tag else meta.get_obj_class(name)
    if item is None:
        item = TagMeta(name, item_type) if is_tag else ObjClass(name, item_type)
        meta = meta.add_tag_meta(item) if is_tag else meta.add_obj_class(item)
        return item, meta
    if is_tag and item.value_type != item_type or not is_tag and item.geometry_type != item_type:
        new_name = generate_new_name_for_meta(meta, name, item_type, is_tag=is_tag)
        if new_name != name:
            logger.warn(
                f"Type mismatch for {'tag' if is_tag else 'class'} '{name}'. Renamed to '{new_name}'"
            )
            item = TagMeta(new_name, item_type) if is_tag else ObjClass(new_name, item_type)
            meta = meta.add_tag_meta(item) if is_tag else meta.add_obj_class(item)
    return item, meta


def get_or_create_tag_meta(meta: ProjectMeta, tag_name: str) -> Tuple[TagMeta, ProjectMeta]:
    """Get or create TagMeta in project meta."""

    return _get_or_create_tag_or_cls(meta, tag_name)


def get_or_create_obj_cls(
    meta: ProjectMeta, cls_name: str, geometry_name: str
) -> Tuple[ObjClass, ProjectMeta]:
    """Get or create ObjClass in project meta."""

    return _get_or_create_tag_or_cls(meta, cls_name, geometry_name)


def create_supervisely_annotation(image_path: str, ann: Dict, meta: ProjectMeta) -> Annotation:
    """Create Supervisely annotation from Label Studio annotation."""

    sly_ann = Annotation.from_img_path(image_path)
    h, w = sly_ann.img_size

    relations = []  # list of relations (from_id, to_id)
    key_label_map = defaultdict(list)  # map for object id to created list of Label objects
    img_tags = []  # list of tags

    objects = ann.get("result", [])
    for obj in objects:
        item_type = obj.get("type")
        if item_type == "relation":
            relations.append((obj["from_id"], obj["to_id"]))
            continue
        item_id = obj.get("id")
        item_data = obj.get("value", {})
        item_name = item_data.get(item_type, [])
        if len(item_name) == 0:
            continue
        item_name = item_name[0]
        if item_type in POSSIBLE_SHAPE_TYPES:
            geom = convert_geometry_to_sly(item_type, item_data, h, w)
            if geom is None:
                continue
            obj_cls, meta = get_or_create_obj_cls(meta, item_name, geom.geometry_name())
            key_label_map[item_id].append(Label(geom, obj_cls))
        elif item_type in POSSIBLE_TAGS_TYPES:
            tag_meta, meta = get_or_create_tag_meta(meta, item_name)
            img_tags.append(Tag(tag_meta))

    res_labels = []
    for from_id, to_id in relations:
        key = uuid4().hex
        labels = key_label_map.pop(from_id, []) + key_label_map.pop(to_id, [])
        res_labels.extend([label.clone(binding_key=key) for label in labels])

    for labels in key_label_map.values():
        res_labels.extend(labels)

    sly_ann = sly_ann.clone(labels=res_labels, img_tags=img_tags)
    return sly_ann, meta



from typing import Callable, Dict, List, Optional, Union

import supervisely.io.env as env
from supervisely.api.api import Api
from supervisely.app.widgets import Widget
from supervisely.app.widgets.container.container import Container
from supervisely.app.widgets.select.select import Select
from supervisely.app.widgets.tree_select.tree_select import TreeSelect
from supervisely.project.project_type import ProjectType


class SelectDatasetTree(Widget):
    """SelectDatasetTree widget in Supervisely is a widget that allows users to select datasets from a tree-like structure.
    Read about it in `Developer Portal <https://developer.supervisely.com/app-development/widgets/selection/selectdatasettree>`_
        (including screenshots and examples).

    :param default_id: The ID of the dataset to be selected by default.
    :type default_id: Union[int, None]
    :param project_id: The ID of the project to read datasets from.
    :type project_id: Union[int, None]
    :param multiselect: Whether multiple datasets can be selected.
    :type multiselect: bool
    :param compact: Whether the widget should be compact (e.g. no team, workspace, and project selectors).
    :type compact: bool
    :param select_all_datasets: Whether all datasets should be selected by default.
    :type select_all_datasets: bool
    :param allowed_project_types: The list of project types that are allowed to be selected.
    :type allowed_project_types: Optional[List[ProjectType]]
    :param flat: Whether the dataset selector should be flat.
    :type flat: bool
    :param always_open: Whether the dataset selector should always be open.
    :type always_open: bool
    :param team_is_selectable: Whether the team selector should be selectable.
    :type team_is_selectable: bool
    :param workspace_is_selectable: Whether the workspace selector should be selectable.
    :type workspace_is_selectable: bool
    :param append_to_body: Determines where the popover is attached. If False, it is positioned inside the input's container. This can cause the popover to be hidden if the input is within a Card or a widget that restricts visibility.
    :type append_to_body: bool
    :param widget_id: The unique identifier of the widget.
    :type widget_id: Union[str, None]

    :Public methods:
    - `set_project_id(project_id: int) -> None`: Set the project ID to read datasets from.
    - `get_selected_ids() -> Optional[List[int]]`: Get the IDs of the selected datasets.
    - `get_selected_id() -> Optional[int]`: Get the ID of the selected dataset.
    - `value_changed(func: Callable) -> Callable`: Decorator to set the callback function for the value changed event.
    - `set_dataset_id(dataset_id: int) -> None`: Set the ID of the dataset to be selected by default.
    - `set_dataset_ids(dataset_ids: List[int]) -> None`: Set the IDs of the datasets to be selected by default.
    - `get_selected_project_id() -> Optional[int]`: Get the ID of the selected project.
    - `get_selected_team_id() -> int`: Get the ID of the selected team.
    - `set_team_id(team_id: int) -> None`: Set the team ID to read workspaces from.
    - `get_selected_workspace_id() -> int`: Get the ID of the selected workspace.
    - `set_workspace_id(workspace_id: int) -> None`: Set the workspace ID to read projects from.
    - `is_all_selected() -> bool`: Check if all datasets are selected.
    - `select_all() -> None`: Select all datasets.

    :Properties:
    - `team_id`: The ID of the team selected in the widget.
    - `workspace_id`: The ID of the workspace selected in the widget.
    - `project_id`: The ID of the project selected in the widget.

    :Usage example:

        .. code-block:: python
            from supervisely.app.widgets import SelectDatasetTree

            project_id = 123
            dataset_id = 456

            select_dataset_tree = SelectDatasetTree(
                default_id=dataset_id,
                project_id=project_id,
                multiselect=True,
                flat=True)

            @select_dataset_tree.value_changed
            def on_change(selected_ids):
                print(selected_ids) # Output: [456, 789]
    """

    def __init__(
        self,
        default_id: Union[int, None] = None,
        project_id: Union[int, None] = None,
        multiselect: bool = False,
        compact: bool = False,
        select_all_datasets: bool = False,
        allowed_project_types: Optional[List[ProjectType]] = None,
        flat: bool = False,
        always_open: bool = False,
        team_is_selectable: bool = True,
        workspace_is_selectable: bool = True,
        append_to_body: bool = True,
        widget_id: Union[str, None] = None,
    ):
        self._api = Api()

        if default_id is not None and project_id is None:
            raise ValueError("Project ID must be provided when default dataset ID is set.")

        if not multiselect and select_all_datasets:
            raise ValueError("Select all datasets is only available in multiselect mode.")

        # Reading team_id and workspace_id from environment variables.
        # If not found, error will be raised.
        self._team_id = env.team_id()
        self._workspace_id = env.workspace_id()

        # Using environment variables to set the default values if they are not provided.
        self._project_id = project_id or env.project_id(raise_not_found=False)
        self._dataset_id = default_id or env.dataset_id(raise_not_found=False)

        self._multiselect = multiselect
        self._compact = compact
        self._append_to_body = append_to_body

        # Extract values from Enum to match the .type property of the ProjectInfo object.
        self._project_types = (
            [project_type.value for project_type in allowed_project_types]
            if allowed_project_types is not None
            else None
        )

        # Widget components.
        self._select_team = None
        self._select_workspace = None
        self._select_project = None
        self._select_dataset = None

        # List of widgets will be used to create a Container.
        self._widgets = []

        if not compact:
            # If the widget is not compact, create team, workspace, and project selectors.
            self._create_selectors(team_is_selectable, workspace_is_selectable)

        # Create the dataset selector.
        self._create_dataset_selector(flat, always_open, select_all_datasets)

        # Group the selectors and the dataset selector into a container.
        self._content = Container(self._widgets)
        super().__init__(widget_id=widget_id, file_path=__file__)

    def disable(self):
        """Disable the widget in the UI."""
        for widget in self._widgets:
            widget.disable()

    def enable(self) -> None:
        """Enable the widget in the UI."""
        for widget in self._widgets:
            widget.enable()

    @property
    def team_id(self) -> int:
        """The ID of the team selected in the widget.

        :return: The ID of the team.
        :rtype: int
        """
        return self._team_id

    @team_id.setter
    def team_id(self, team_id: int) -> None:
        """Set the team ID to read workspaces from.

        :param team_id: The ID of the team.
        :type team_id: int
        """
        if not self._compact:
            self._select_team.set_value(team_id)
            self._select_workspace.set(self._get_select_items(team_id=team_id))
        self._team_id = team_id

    def get_selected_team_id(self) -> int:
        """Get the ID of the selected team.

        :return: The ID of the selected team.
        :rtype: int
        """
        return self.team_id

    def set_team_id(self, team_id: int) -> None:
        """Set the team ID to read workspaces from.

        :param team_id: The ID of the team.
        :type team_id: int
        """
        self.team_id = team_id

    @property
    def workspace_id(self) -> int:
        """The ID of the workspace selected in the widget.

        :return: The ID of the workspace.
        :rtype: int
        """
        return self._workspace_id

    @workspace_id.setter
    def workspace_id(self, workspace_id: int) -> None:
        """Set the workspace ID to read projects from.

        :param workspace_id: The ID of the workspace.
        :type workspace_id: int
        """
        if not self._compact:
            self._select_workspace.set_value(workspace_id)
            self._select_project.set(self._get_select_items(workspace_id=workspace_id))
        self._workspace_id = workspace_id

    def get_selected_workspace_id(self) -> int:
        """Get the ID of the selected workspace.

        :return: The ID of the selected workspace.
        :rtype: int
        """
        return self.workspace_id

    def set_workspace_id(self, workspace_id: int) -> None:
        """Set the workspace ID to read projects from.

        :param workspace_id: The ID of the workspace.
        :type workspace_id: int
        """
        self.workspace_id = workspace_id

    @property
    def project_id(self) -> Optional[int]:
        """The ID of the project selected in the widget.

        :return: The ID of the project.
        :rtype: Optional[int]
        """
        return self._project_id

    @project_id.setter
    def project_id(self, project_id: Optional[int]) -> None:
        """Set the project ID to read datasets from.

        :param project_id: The ID of the project.
        :type project_id: int
        """
        if not self._compact:
            self._select_project.set_value(project_id)
        self._project_id = project_id
        self._select_dataset.set_items(self._read_datasets(project_id))

    def get_selected_project_id(self) -> Optional[int]:
        """Get the ID of the selected project.

        :return: The ID of the selected project.
        :rtype: Optional[int]
        """
        return self.project_id

    def set_dataset_id(self, dataset_id: int) -> None:
        """Set the ID of the dataset to be selected by default.

        :param id: The ID of the dataset.
        :type id: int
        """
        self._select_dataset.set_selected_by_id(dataset_id)

    def set_dataset_ids(self, dataset_ids: List[int]) -> None:
        """Set the IDs of the datasets to be selected by default.

        :raise ValueError: If multiselect is disabled.
        :param ids: The IDs of the datasets.
        :type ids: List[int]
        """
        if not self._multiselect:
            raise ValueError("This method can only be called when multiselect is enabled.")
        self._select_dataset.set_selected_by_id(dataset_ids)

    def value_changed(self, func: Callable) -> Callable:
        """Decorator to set the callback function for the value changed event.

        :param func: The callback function.
        :type func: Callable
        :return: The callback function.
        :rtype: Callable
        """

        @self._select_dataset.value_changed
        def _click(items: Union[List[TreeSelect.Item], TreeSelect.Item]):
            if isinstance(items, list):
                res = [item.id for item in items]
            else:
                res = items.id

            func(res)

        return _click

    def _create_dataset_selector(
        self, flat: bool, always_open: bool, select_all_datasets: bool
    ) -> None:
        """Create the dataset selector.

        :param flat: Whether the dataset selector should be flat.
        :type flat: bool
        :param always_open: Whether the dataset selector should always be open.
        :type always_open: bool
        :param select_all_datasets: Whether all datasets should be selected by default.
        :type select_all_datasets: bool
        """
        self._select_dataset = TreeSelect(
            items=self._read_datasets(self._project_id),
            multiple_select=self._multiselect,
            flat=flat,
            always_open=always_open,
            append_to_body=self._append_to_body,
            width=193,
            placeholder="Select dataset",
        )
        if self._dataset_id is not None:
            self._select_dataset.set_selected_by_id(self._dataset_id)
        if select_all_datasets:
            self._select_dataset.select_all()

        # Adding the dataset selector to the list of widgets to be added to the container.
        self._widgets.append(self._select_dataset)

    def _create_selectors(self, team_is_selectable: bool, workspace_is_selectable: bool):
        """Create the team, workspace, and project selectors.

        :param team_is_selectable: Whether the team selector should be selectable.
        :type team_is_selectable: bool
        :param workspace_is_selectable: Whether the workspace selector should be selectable.
        :type workspace_is_selectable: bool
        """

        def team_selector_handler(team_id: int):
            """Handler function for the event when the team selector value changes.

            :param team_id: The ID of the selected team.
            :type team_id: int
            """
            self._select_workspace.set(items=self._get_select_items(team_id=team_id))
            self._team_id = team_id

        def workspace_selector_handler(workspace_id: int):
            """Handler function for the event when the workspace selector value changes.

            :param workspace_id: The ID of the selected workspace.
            :type workspace_id: int
            """
            self._select_project.set(items=self._get_select_items(workspace_id=workspace_id))
            self._workspace_id = workspace_id

        def project_selector_handler(project_id: int):
            """Handler function for the event when the project selector value changes.

            :param project_id: The ID of the selected project.
            :type project_id: int
            """
            self._select_dataset.set_items(self._read_datasets(project_id))
            self._project_id = project_id

        self._select_team = Select(
            items=self._get_select_items(), placeholder="Select team", filterable=True
        )
        self._select_team.set_value(self._team_id)
        if not team_is_selectable:
            self._select_team.disable()

        self._select_workspace = Select(
            items=self._get_select_items(team_id=self._team_id),
            placeholder="Select workspace",
            filterable=True,
        )
        self._select_workspace.set_value(self._workspace_id)
        if not workspace_is_selectable:
            self._select_workspace.disable()

        self._select_project = Select(
            items=self._get_select_items(workspace_id=self._workspace_id),
            placeholder="Select project",
            filterable=True,
        )
        self._select_project.set_value(self._project_id)

        # Register the event handlers.
        self._select_team.value_changed(team_selector_handler)
        self._select_workspace.value_changed(workspace_selector_handler)
        self._select_project.value_changed(project_selector_handler)

        # Adding widgets to the list, so they can be added to the container.
        self._widgets.extend([self._select_team, self._select_workspace, self._select_project])

    def _get_select_items(self, **kwargs) -> List[Select.Item]:
        """Get the list of items for the team, workspace, and project selectors.
        Possible keyword arguments are 'team_id' and 'workspace_id'.

        :return: The list of items.
        :rtype: List[Select.Item]
        """
        if not kwargs:
            items = self._api.team.get_list()
        elif "team_id" in kwargs:
            items = self._api.workspace.get_list(kwargs["team_id"])
        elif "workspace_id" in kwargs:
            projects_list = self._api.project.get_list(kwargs["workspace_id"])
            if self._project_types is not None:
                items = [
                    project for project in projects_list if project.type in self._project_types
                ]  # TODO: Filter project from API, not from here.
            else:
                items = projects_list

        return [Select.Item(value=item.id, label=item.name) for item in items]

    def get_json_data(self) -> Dict:
        """Get the JSON data of the widget.

        :return: The JSON data.
        :rtype: Dict
        """
        return {}

    def get_json_state(self) -> Dict:
        """Get the JSON state of the widget.

        :return: The JSON state.
        :rtype: Dict
        """
        return {}

    def _read_datasets(self, project_id: Optional[int]) -> Optional[List[TreeSelect.Item]]:
        """Get the lisf of TreeSelect.Item objects representing the dataset hierarchy.

        :param project_id: The ID of the project.
        :type project_id: Optional[int]
        :return: The list of TreeSelect.Item objects.
        :rtype: Optional[List[TreeSelect.Item]]
        """
        if not project_id:
            return None
        dataset_tree = self._api.dataset.get_tree(project_id)

        def convert_tree_to_list(node, parent_id: Optional[int] = None):
            """
            Recursively converts a tree of DatasetInfo objects into a list of
                SelectDatasetTree.Item objects.

            :param node: The current node in the tree (a tuple of DatasetInfo and its children).
            :param parent_id: The ID of the parent dataset, if any.
            :return: A list of SelectDatasetTree.Item objects representing the tree.
            """
            result = []
            for dataset_info, children in node.items():
                item = TreeSelect.Item(
                    id=dataset_info.id,
                    label=dataset_info.name,
                    children=convert_tree_to_list(children, parent_id=dataset_info.id),
                )

                result.append(item)

            return result

        return convert_tree_to_list(dataset_tree)

    def set_project_id(self, project_id: Optional[int]) -> None:
        """Set the project ID to read datasets from.

        :param project_id: The ID of the project.
        :type project_id: int
        """
        self.project_id = project_id

    def _get_selected(self) -> Optional[Union[List[int], int]]:
        """Get the ID of the selected dataset(s).

        :return: The ID of the selected dataset(s).
        :rtype: Optional[Union[List[int], int]]
        """
        selected = self._select_dataset.get_selected()
        if not selected:
            return None

        if isinstance(selected, list):
            return [item.id for item in selected]
        else:
            return selected.id

    def get_selected_ids(self) -> Optional[List[int]]:
        """Get the IDs of the selected datasets.

        raise ValueError if multiselect is disabled.
        return: The IDs of the selected datasets.
        rtype: Optional[List[int]]
        """
        if not self._multiselect:
            raise ValueError("This method can only be called when multiselect is enabled.")
        return self._get_selected()

    def get_selected_id(self) -> Optional[int]:
        """Get the ID of the selected dataset.

        raise ValueError if multiselect is enabled.
        return: The ID of the selected dataset.
        rtype: Optional[int]
        """
        if self._multiselect:
            raise ValueError("This method can only be called when multiselect is disabled.")
        return self._get_selected()

    def is_all_selected(self) -> bool:
        """Check if all datasets are selected.

        return: True if all datasets are selected, False otherwise.
        rtype: bool
        """
        return self._select_dataset.is_all_selected()

    def select_all(self) -> None:
        """Select all datasets."""
        self._select_dataset.select_all()



# coding: utf-8
"""utilities used for labeling jobs"""

from __future__ import annotations

import urllib.parse
from typing import List, NamedTuple, Tuple

from supervisely.api.labeling_job_api import LabelingJobApi

Status = LabelingJobApi.Status
from pandas import DataFrame  # for typehints

from supervisely.api.module_api import ApiField
from supervisely.labeling_jobs import constants as constants


def total_desc() -> Tuple[str, str]:
    """
    :return: Description about total number of Labeling Jobs
    :rtype: :class:`Tuple[str, str]`
    """
    return "TOTAL", "the total number of jobs in current team"


def is_completed_desc() -> Tuple[str, str]:
    """
    :return: Description about total number of completed Labeling Jobs
    :rtype: :class:`Tuple[str, str]`
    """
    return "COMPLETED", "the number of completed jobs"


def is_completed(job_info: NamedTuple) -> bool:
    """
    :return: True if Labeling Job is completed, otherwise None
    :rtype: :class:`bool` or :class:`None`
    """
    if job_info.status == str(Status.COMPLETED):
        return True


def is_stopped_desc() -> Tuple[str, str]:
    """
    :return: Description about number of stopped Labeling Jobs
    :rtype: :class:`Tuple[str, str]`
    """
    return "STOPPED", "the number of stopped jobs"


def is_stopped(job_info: NamedTuple) -> bool:
    """
    :return: True if Labeling Job is stopped, otherwise None
    :rtype: :class:`bool` or :class:`None`
    """
    if job_info.status == str(Status.STOPPED):
        return True


def is_not_started_desc() -> Tuple[str, str]:
    """
    :return: Description about total number of pending Labeling Jobs
    :rtype: :class:`Tuple[str, str]`
    """
    return "PENDING", "the number of jobs labeler haven't even opened yet (created but not started)"


def is_not_started(job_info: NamedTuple) -> bool:
    """
    :return: True if Labeling Job is not started, otherwise None
    :rtype: :class:`bool` or :class:`None`
    """
    if job_info.status == str(Status.PENDING):
        return True


def total_items_count_desc() -> Tuple[str, str]:
    """
    :return: Description about total number of items in all Labeling Jobs
    :rtype: :class:`Tuple[str, str]`
    """
    return "TOTAL", "the total number of items in all labeling jobs"


def total_items_count(job_info: NamedTuple) -> int:
    """
    :return: Number of total items count in Labeling Jobs
    :rtype: :class:`int`
    """
    return job_info.images_count


def labeled_items_count_desc() -> Tuple[str, str]:
    """
    :return: Description about total number of labeled items in all Labeling Jobs
    :rtype: :class:`Tuple[str, str]`
    """
    return (
        "LABELED",
        'the total number of labeled items (labelers marked as "finished") in all labeling jobs',
    )


# cnt images, that labeler marked as done
def labeled_items_count(job_info: NamedTuple) -> int:
    """
    :return: Number of Images, that labeler marked as done
    :rtype: :class:`int`
    """
    if is_on_labeling(job_info):
        return job_info.finished_images_count
    else:
        return total_items_count(job_info)


def reviewed_items_count_desc() -> Tuple[str, str]:
    """
    :return: Description about total number of reviewed items in all Labeling Jobs
    :rtype: :class:`Tuple[str, str]`
    """
    return (
        "REVIEWED",
        'the total number of reviewed items (reviewers marked as "accepted" or "rejected") in all labeling jobs',
    )


# cnt images, that reviewer accepted or rejected
def reviewed_items_count(job_info: NamedTuple) -> int:
    """
    :return: Number of reviewed Images(accepted and rejected)
    :rtype: :class:`int`
    """
    return job_info.accepted_images_count + job_info.rejected_images_count


def accepted_items_count_desc() -> Tuple[str, str]:
    """
    :return: Description about total number of accepted items in all Labeling Jobs
    :rtype: :class:`Tuple[str, str]`
    """
    return (
        "ACCEPTED",
        'the total number of accepted items (reviewers marked as "accepted") in all labeling jobs',
    )


def accepted_items_count(job_info: NamedTuple) -> int:
    """
    :return: Number of accepted images in all Labeling Jobs
    :rtype: :class:`int`
    """
    return job_info.accepted_images_count


def rejected_items_count_desc() -> Tuple[str, str]:
    """
    :return: Description about total number of rejected items in Labeling Jobs
    :rtype: :class:`Tuple[str, str]`
    """
    return (
        "REJECTED",
        'the total number of rejected items (reviewers marked as "rejected") in all labeling jobs',
    )


def rejected_items_count(job_info: NamedTuple) -> int:
    """
    :return: Number of rejected images in all Labeling Jobs
    :rtype: :class:`int`
    """
    return job_info.rejected_images_count


def is_on_labeling_desc() -> str:
    return "LABELING IN PROGRESS", "the number of jobs with status IN_PROGRESS"


# labeling is in progress
def is_on_labeling(job_info: NamedTuple) -> bool:
    """
    :return: True if Labeling Job is in progress, False otherwise
    :rtype: :class:`bool`
    """
    if job_info.status == str(Status.IN_PROGRESS):
        return True
    return False


def is_labeling_started_desc() -> Tuple[str, str]:
    """
    :return: Description about total number of Labeling Jobs that are started by labeler
    :rtype: :class:`Tuple[str, str]`
    """
    return (
        "LABELING STARTED",
        'the number of jobs that are started by labeler and with at least one labeled item (marked "done" by labeler)',
    )


# cnt jobs that are started by labeler and with at least one image that marked "done" by labeler
def is_labeling_started(job_info: NamedTuple) -> bool:
    """
    :return: True if Labeling Job is started, False otherwise
    :rtype: :class:`bool`
    """
    if is_on_labeling(job_info) and labeled_items_count(job_info) != 0:
        return True
    return False


def is_on_review_desc() -> Tuple[str, str]:
    """
    :return: Description about number of Labeling Jobs with status 'ON REVIEW'
    :rtype: :class:`Tuple[str, str]`
    """
    return "ON REVIEW", "the number of jobs with status ON_REVIEW"


# cnt jobs with "on review" status
def is_on_review(job_info: NamedTuple) -> bool:
    """
    :return: True if Labeling Job is in 'ON REVIEW' status , False otherwise
    :rtype: :class:`bool`
    """
    if job_info.status == str(Status.ON_REVIEW):
        return True
    return False


def is_review_started_desc() -> Tuple[str, str]:
    """
    :return: Description about number of Labeling Jobs that are started by reviewer
    :rtype: :class:`Tuple[str, str]`
    """
    return (
        "REVIEW STARTED",
        'the number of jobs that are started by reviewer - with at least one reviewed item (marked "accepted" or "rejected")',
    )


# cnt jobs with at least one reviewed (accepted or rejected) item
def is_review_started(job_info: NamedTuple) -> bool:
    """
    :return: True if Labeling Job is in 'review' status and there are Images that reviewer accepted or rejected, False otherwise
    :rtype: :class:`bool`
    """
    if is_on_review(job_info) and reviewed_items_count(job_info) != 0:
        return True
    return False


def is_zero_labeling_desc() -> Tuple[str, str]:
    """
    :return: Description about number of Labeling Jobs with status "IN PROGRESS" with zero labeled items
    :rtype: :class:`Tuple[str, str]`
    """
    return "ZERO LABELED", 'the number of jobs with status "IN PROGRESS" with zero labeled items'


def is_zero_reviewed_desc() -> Tuple[str, str]:
    """
    :return: Description about number of Labeling Jobs with status "ON REVIEW" with zero reviewed items
    :rtype: :class:`Tuple[str, str]`
    """
    return "ZERO REVIEWED", 'the number of jobs with status "ON REVIEW" with zero reviewed items'


def get_job_url(server_address: str, job: NamedTuple) -> str:
    """
    Get labeling job url.

    :param server_address: Server address there labeling job working.
    :type server_address: str
    :param job: Information about labeling job.
    :type job: NamedTuple
    :return: Labeling job url
    :rtype: :class:`str`
    :Usage example:

     .. code-block:: python

        address = 'https://app.supervise.ly'
        os.environ['SERVER_ADDRESS'] = address
        os.environ['API_TOKEN'] = 'Your Supervisely API Token'
        api = sly.Api.from_env()
        job_info = api.labeling_job.get_info_by_id(2)

        from supervisely.labeling_jobs.utils import get_job_url
        job_url = get_job_url(address, job_info)
        print(job_url)
        # Output: https://app.supervise.ly/app/images/4/8/58/54?jobId=2
    """
    result = urllib.parse.urljoin(
        server_address,
        "app/images/{}/{}/{}/{}?jobId={}".format(
            job.team_id, job.workspace_id, job.project_id, job.dataset_id, job.id
        ),
    )

    return result


def jobs_stats(server_address: str, jobs: List[NamedTuple], stats: List) -> DataFrame:
    """
    Get statistics about Labeling Jobs items.

    :param server_address: Supervisely Server Address.
    :type server_address: str
    :param jobs: List of information about Labeling Jobs.
    :type jobs: List[NamedTuple]
    :param stats:
    :type stats: list
    :return: Statistics about Labeling Jobs items as `pd.DataFrame <https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.html>`_
    :rtype: :class:`pd.DataFrame`
    :Usage example:

     .. code-block:: python

        os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
        os.environ['API_TOKEN'] = 'Your Supervisely API Token'
        api = sly.Api.from_env()

        job_info1 = api.labeling_job.get_info_by_id(2)
        job_info2 = api.labeling_job.get_info_by_id(3)
        address = https://app.supervise.ly

        from supervisely.labeling_jobs.utils import jobs_stats
        result = jobs_stats(address, [job_info1, job_info2], [1, 2])
        print(result)
        # Output:
        #    ID                                               NAME     STATUS  TOTAL  LABELED  REVIEWED  ACCEPTED  REJECTED        CREATED_AT
        # 0   2  <a href="https://app.supervise.ly/app/images/...  completed      3        3         3         2         1  08/04/2020 15:10
        # 1   3  <a href="https://app.supervise.ly/app/images/...  completed      2        2         2         2         0  08/04/2020 15:10
    """
    import pandas as pd

    col_job_id = []
    col_job_name = []  # link here
    col_job_status = []
    col_items_total = []
    col_items_labeled = []
    col_items_reviewed = []
    col_items_accepted = []
    col_items_rejected = []
    col_created_at = []

    for job, stat in zip(jobs, stats):
        col_job_id.append(job.id)
        col_job_name.append(
            '<a href="{0}" rel="noopener noreferrer" target="_blank">{1}</a>'.format(
                get_job_url(server_address, job), job.name
            )
        )
        col_job_status.append(job.status)
        col_created_at.append(job.created_at)
        col_items_total.append(total_items_count(job))
        col_items_labeled.append(labeled_items_count(job))
        col_items_reviewed.append(reviewed_items_count(job))
        col_items_accepted.append(accepted_items_count(job))
        col_items_rejected.append(rejected_items_count(job))

    df = pd.DataFrame(
        list(
            zip(
                col_job_id,
                col_job_name,
                col_job_status,
                col_items_total,
                col_items_labeled,
                col_items_reviewed,
                col_items_accepted,
                col_items_rejected,
                col_created_at,
            )
        ),
        columns=[
            "ID",
            "NAME",
            "STATUS",
            "TOTAL",
            "LABELED",
            "REVIEWED",
            "ACCEPTED",
            "REJECTED",
            "CREATED_AT",
        ],
    )

    df["CREATED_AT"] = pd.to_datetime(df["CREATED_AT"]).dt.strftime("%d/%m/%Y %H:%M")
    return df


def jobs_summary(jobs: List[NamedTuple]) -> DataFrame:
    """
    Get summary statistics about given Labeling Jobs.

    :param jobs: List of information about Labeling Jobs.
    :type jobs: List[NamedTuple]
    :return: Statistics about Labeling Jobs as `pd.DataFrame <https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.html>`_
    :rtype: :class:`pd.DataFrame`
    :Usage example:

     .. code-block:: python

        os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
        os.environ['API_TOKEN'] = 'Your Supervisely API Token'
        api = sly.Api.from_env()

        job_info1 = api.labeling_job.get_info_by_id(2)
        job_info2 = api.labeling_job.get_info_by_id(3)

        from supervisely.labeling_jobs.utils import jobs_summary
        result = jobs_summary([job_info1, job_info2])
        print(result)
        # Output:
        #                 JOB STATUS  QUANTITY PERCENTAGE                                        DESCRIPTION
        # 0  0                 TOTAL         2   100.00 %           the total number of jobs in current team
        # 1  1             COMPLETED         2   100.00 %                       the number of completed jobs
        # 2  2               STOPPED         0     0.00 %                         the number of stopped jobs
        # 3  3               PENDING         0     0.00 %  the number of jobs labeler haven't even opened...
        # 4  4  LABELING IN PROGRESS         0     0.00 %         the number of jobs with status IN_PROGRESS
        # 5  5      LABELING STARTED         0     0.00 %  the number of jobs that are started by labeler...
        # 6  6          ZERO LABELED         0     0.00 %  the number of jobs with status "IN PROGRESS" w...
        # 7  7             ON REVIEW         0     0.00 %           the number of jobs with status ON_REVIEW
        # 8  8        REVIEW STARTED         0     0.00 %  the number of jobs that are started by reviewe...
        # 9  9         ZERO REVIEWED         0     0.00 %  the number of jobs with status "ON REVIEW" wit...
    """
    import pandas as pd

    count_total = len(jobs)

    count_completed = 0
    count_stopped = 0

    count_labeling_not_started = 0  # pending

    count_on_labeling = 0
    count_labeling_started = 0
    count_labeling_zero_done = 0

    count_on_review = 0
    count_review_started = 0
    count_review_zero_done = 0

    for job in jobs:
        if is_completed(job):
            count_completed += 1
        elif is_stopped(job):
            count_stopped += 1
        elif is_not_started(job):
            count_labeling_not_started += 1
        elif is_on_labeling(job):
            count_on_labeling += 1
            if is_labeling_started(job):
                count_labeling_started += 1
        elif is_on_review(job):
            count_on_review += 1
            if is_review_started(job):
                count_review_started += 1
        else:
            raise RuntimeError("Unhandled job status: {}".format(str(job)))

    count_labeling_zero_done = count_on_labeling - count_labeling_started
    count_review_zero_done = count_on_review - count_review_started

    names = []
    percentages = []
    descriptions = []

    table_rows_f = [
        total_desc,
        is_completed_desc,
        is_stopped_desc,
        is_not_started_desc,
        is_on_labeling_desc,
        is_labeling_started_desc,
        is_zero_labeling_desc,
        is_on_review_desc,
        is_review_started_desc,
        is_zero_reviewed_desc,
    ]

    values = [
        count_total,
        count_completed,
        count_stopped,
        count_labeling_not_started,
        count_on_labeling,
        count_labeling_started,
        count_labeling_zero_done,
        count_on_review,
        count_review_started,
        count_review_zero_done,
    ]

    for v, func in zip(values, table_rows_f):
        percentages.append("{:.2f} %".format(v * 100 / count_total))
        name, desc = func()
        names.append(name)
        descriptions.append(desc)

    df = pd.DataFrame(
        list(zip(list(range(len(names))), names, values, percentages, descriptions)),
        columns=["#", "JOB STATUS", "QUANTITY", "PERCENTAGE", "DESCRIPTION"],
    )
    return df


def images_summary(jobs: List[NamedTuple]) -> DataFrame:
    """
    Get summary statistics about given Labeling Jobs images.

    :param jobs: List of information about Labeling Jobs.
    :type jobs: List[NamedTuple]
    :return: Statistics about Labeling Jobs images
    :rtype: :class:`pd.DataFrame`
    :Usage example:

     .. code-block:: python

        os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
        os.environ['API_TOKEN'] = 'Your Supervisely API Token'
        api = sly.Api.from_env()

        job_info1 = api.labeling_job.get_info_by_id(2)
        job_info2 = api.labeling_job.get_info_by_id(3)

        from supervisely.labeling_jobs.utils import images_summary
        result = images_summary([job_info1, job_info2])
        print(result)
        # Output:
        #    # ITEM STATUS  QUANTITY PERCENTAGE                                        DESCRIPTION
        # 0  0       TOTAL         5   100.00 %     the total number of items in all labeling jobs
        # 1  1     LABELED         5   100.00 %  the total number of labeled items (labelers ma...
        # 2  2    REVIEWED         5   100.00 %  the total number of reviewed items (reviewers ...
        # 3  3    ACCEPTED         4    80.00 %  the total number of accepted items (reviewers ...
        # 4  4    REJECTED         1    20.00 %  the total number of rejected items (reviewers ...
    """
    import pandas as pd

    count_total_items = 0
    count_labeled_items = 0
    count_reviewed_items = 0
    count_accepted_items = 0
    count_rejected_items = 0

    for job in jobs:
        count_total_items += total_items_count(job)
        count_labeled_items += labeled_items_count(job)
        count_reviewed_items += reviewed_items_count(job)
        count_accepted_items += accepted_items_count(job)
        count_rejected_items += rejected_items_count(job)

    values_items = [
        count_total_items,
        count_labeled_items,
        count_reviewed_items,
        count_accepted_items,
        count_rejected_items,
    ]

    items_f = [
        total_items_count_desc,
        labeled_items_count_desc,
        reviewed_items_count_desc,
        accepted_items_count_desc,
        rejected_items_count_desc,
    ]

    names_items = []
    percentages_items = []
    descriptions_items = []

    for v, func in zip(values_items, items_f):
        percentages_items.append("{:.2f} %".format(v * 100 / count_total_items))
        name, desc = func()
        names_items.append(name)
        descriptions_items.append(desc)

    df = pd.DataFrame(
        list(
            zip(
                list(range(len(names_items))),
                names_items,
                values_items,
                percentages_items,
                descriptions_items,
            )
        ),
        columns=["#", "ITEM STATUS", "QUANTITY", "PERCENTAGE", "DESCRIPTION"],
    )
    return df


def classes_summary(stats: List) -> DataFrame:
    import pandas as pd

    class_id_stats = {}
    for stat in stats:
        for class_stat in stat[constants.CLASSES_STATS]:
            class_id = class_stat[ApiField.ID]
            class_name = class_stat[ApiField.NAME]
            class_shape = class_stat[ApiField.SHAPE]
            image_count = class_stat[ApiField.IMAGES_COUNT]
            object_count = class_stat[ApiField.LABELS_COUNT]
            class_color = class_stat[ApiField.COLOR]

            if class_id not in class_id_stats:
                class_id_stats[class_id] = {
                    ApiField.NAME: class_name,
                    ApiField.SHAPE: class_shape,
                    ApiField.COLOR: class_color,
                    ApiField.IMAGES_COUNT: 0,
                    ApiField.LABELS_COUNT: 0,
                }

            class_id_stats[class_id][ApiField.IMAGES_COUNT] += image_count
            class_id_stats[class_id][ApiField.LABELS_COUNT] += object_count

    col_name = []
    col_shape = []
    col_image_count = []
    col_object_count = []
    for class_id, value in class_id_stats.items():
        col_name.append(
            '<b style="display: inline-block; border-radius: 50%; background: {}; width: 8px; height: 8px"></b> {}'.format(
                value[ApiField.COLOR], value[ApiField.NAME]
            )
        )
        col_shape.append(value[ApiField.SHAPE])
        col_image_count.append(value[ApiField.IMAGES_COUNT])
        col_object_count.append(value[ApiField.LABELS_COUNT])

    df = pd.DataFrame(
        list(
            zip(list(range(len(col_name))), col_name, col_shape, col_image_count, col_object_count)
        ),
        columns=["#", "CLASS", "SHAPE", "IMAGES COUNT", "OBJECTS COUNT"],
    )
    return df


def tags_summary(stats: List) -> DataFrame:
    import pandas as pd

    tag_id_stats = {}
    for stat in stats:
        for tag_stat in stat["job"][constants.TAGS_STATS]:
            if "parentId" in tag_stat:
                # @TODO: consider parentId (for oneOf cases)
                continue
            tag_id = tag_stat[ApiField.ID]
            tag_name = tag_stat[ApiField.NAME]
            image_count = tag_stat[ApiField.IMAGES]
            object_count = tag_stat[ApiField.FIGURES]
            class_color = tag_stat[ApiField.COLOR]

            if tag_id not in tag_id_stats:
                tag_id_stats[tag_id] = {
                    ApiField.NAME: tag_name,
                    ApiField.COLOR: class_color,
                    ApiField.IMAGES_COUNT: 0,
                    ApiField.LABELS_COUNT: 0,
                }

            tag_id_stats[tag_id][ApiField.IMAGES_COUNT] += image_count
            tag_id_stats[tag_id][ApiField.LABELS_COUNT] += object_count

    col_name = []
    col_image_count = []
    col_object_count = []
    for tag_id, value in tag_id_stats.items():
        col_name.append(
            '<b style="display: inline-block; border-radius: 50%; background: {}; width: 8px; height: 8px"></b> {}'.format(
                value[ApiField.COLOR], value[ApiField.NAME]
            )
        )
        col_image_count.append(value[ApiField.IMAGES_COUNT])
        col_object_count.append(value[ApiField.LABELS_COUNT])

    df = pd.DataFrame(
        list(zip(list(range(len(col_name))), col_name, col_image_count, col_object_count)),
        columns=["#", "TAG", "IMAGES COUNT", "OBJECTS COUNT"],
    )
    return df



# coding: utf-8


# docs

from __future__ import annotations
import cv2
import numpy as np
from typing import List, Dict, Optional, Union, Tuple
from supervisely.geometry.point import PointLocation
from supervisely.geometry.rectangle import Rectangle

from shapely.geometry import mapping, LineString, Polygon as ShapelyPolygon
from supervisely.geometry.conversions import shapely_figure_to_coords_list
from supervisely.geometry.point_location import row_col_list_to_points
from supervisely.geometry.vector_geometry import VectorGeometry
from supervisely.geometry.constants import (
    EXTERIOR,
    POINTS,
    LABELER_LOGIN,
    UPDATED_AT,
    CREATED_AT,
    ID,
    CLASS_ID,
)
from supervisely.geometry import validation
from supervisely import logger


class Polyline(VectorGeometry):
    """
    Polyline geometry for a single :class:`Label<supervisely.annotation.label.Label>`. :class:`Polyline<Polyline>` class object is immutable.

    :param exterior: List of exterior coordinates, the Polyline is defined with these points.
    :type exterior: List[PointLocation], List[List[int, int]], List[Tuple[int, int]
    :param sly_id: Polyline ID in Supervisely server.
    :type sly_id: int, optional
    :param class_id: ID of :class:`ObjClass<supervisely.annotation.obj_class.ObjClass>` to which Polyline belongs.
    :type class_id: int, optional
    :param labeler_login: Login of the user who created Polyline.
    :type labeler_login: str, optional
    :param updated_at: Date and Time when Polyline was modified last. Date Format: Year:Month:Day:Hour:Minute:Seconds. Example: '2021-01-22T19:37:50.158Z'.
    :type updated_at: str, optional
    :param created_at: Date and Time when Polyline was created. Date Format is the same as in "updated_at" parameter.
    :type created_at: str, optional
    :raises: :class:`ValueError`, field exterior must contain at least two points to create Polyline object

    :Usage example:

     .. code-block:: python

        import supervisely as sly

        exterior = [sly.PointLocation(730, 2104), sly.PointLocation(2479, 402), sly.PointLocation(1500, 780)]
        # or exterior = [[730, 2104], [2479, 402], [1500, 780]]
        # or exterior = [(730, 2104), (2479, 402), (1500, 780)]
        figure = sly.Polyline(exterior)
    """

    @staticmethod
    def geometry_name():
        """
        """
        return "line"

    def __init__(
        self,
        exterior: Union[
            List[PointLocation], List[List[int, int]], List[Tuple[int, int]]
        ],
        sly_id: Optional[int] = None,
        class_id: Optional[int] = None,
        labeler_login: Optional[int] = None,
        updated_at: Optional[str] = None,
        created_at: Optional[str] = None,
    ):
        if len(exterior) < 2:
            raise ValueError(f'"{EXTERIOR}" field must contain at least two points to create "Polyline" object.')

        super().__init__(
            exterior=exterior,
            interior=[],
            sly_id=sly_id,
            class_id=class_id,
            labeler_login=labeler_login,
            updated_at=updated_at,
            created_at=created_at,
        )

    @classmethod
    def from_json(cls, data: Dict) -> Polyline:
        """
        Convert a json dict to Polyline. Read more about `Supervisely format <https://docs.supervise.ly/data-organization/00_ann_format_navi>`_.

        :param data: Polyline in json format as a dict.
        :type data: dict
        :return: Polyline object
        :rtype: :class:`Polyline<Polyline>`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            figure_json = {
                "points": {
                    "exterior": [
                        [2104, 730],
                        [402, 2479],
                        [780, 1500]
                    ],
                    "interior": []
                }
            }
            figure = sly.Polyline.from_json(figure_json)
        """
        validation.validate_geometry_points_fields(data)
        labeler_login = data.get(LABELER_LOGIN, None)
        updated_at = data.get(UPDATED_AT, None)
        created_at = data.get(CREATED_AT, None)
        sly_id = data.get(ID, None)
        class_id = data.get(CLASS_ID, None)
        return cls(
            exterior=row_col_list_to_points(
                data[POINTS][EXTERIOR], flip_row_col_order=True
            ),
            sly_id=sly_id,
            class_id=class_id,
            labeler_login=labeler_login,
            updated_at=updated_at,
            created_at=created_at,
        )

    def crop(self, rect: Rectangle) -> List[Polyline]:
        """
        Crops current Polyline.

        :param rect: Rectangle object for crop.
        :type rect: Rectangle
        :return: List of Polyline objects
        :rtype: :class:`List[Polyline]<Polyline>`

        :Usage Example:

         .. code-block:: python

            import supervisely as sly

            crop_figures = figure.crop(sly.Rectangle(0, 0, 100, 200))
        """
        try:
            clipping_window = [
                [rect.top, rect.left],
                [rect.top, rect.right],
                [rect.bottom, rect.right],
                [rect.bottom, rect.left],
            ]
            clipping_window_shpl = ShapelyPolygon(clipping_window)

            exterior = self.exterior_np
            intersections_polygon = LineString(exterior).intersection(
                clipping_window_shpl
            )
            mapping_shpl = mapping(intersections_polygon)
        except Exception:
            logger.warn("Line cropping exception, shapely.", exc_info=False)
            raise

        res_lines_pts = shapely_figure_to_coords_list(mapping_shpl)

        # tiny hack to combine consecutive segments
        lines_combined = []
        if res_lines_pts != [()]:
            for simple_l in res_lines_pts:
                if len(lines_combined) > 0:
                    prev = lines_combined[-1]
                    if prev[-1] == simple_l[0]:
                        lines_combined[-1] = list(prev) + list(simple_l[1:])
                        continue
                lines_combined.append(simple_l)

        return [Polyline(row_col_list_to_points(line)) for line in lines_combined]

    def _draw_impl(self, bitmap: np.ndarray, color, thickness=1, config=None):
        """
        """
        self._draw_contour_impl(bitmap, color, thickness, config=config)

    def _draw_contour_impl(self, bitmap: np.ndarray, color, thickness=1, config=None):
        """
        """
        exterior = self.exterior_np[:, ::-1]
        cv2.polylines(
            bitmap, pts=[exterior], isClosed=False, color=color, thickness=thickness
        )

    @property
    def area(self) -> float:
        """
        Polyline area, always 0.0.

        :return: Area of current Polyline
        :rtype: :class:`float`

        :Usage Example:

         .. code-block:: python

            print(figure.area)
            # Output: 0.0
        """
        return 0.0

    def approx_dp(self, epsilon: float) -> Polyline:
        """
        Approximates a Polyline with the specified precision.

        :param epsilon: Specifying the approximation accuracy.
        :type epsilon: float
        :return: Polyline object
        :rtype: :class:`Polyline<Polyline>`

        :Usage Example:

         .. code-block:: python

            # Remember that Polyline class object is immutable, and we need to assign new instance of Polyline to a new variable
            approx_figure = figure.approx_dp(0.75)
        """
        exterior_np = self._approx_ring_dp(
            self.exterior_np, epsilon, closed=True
        ).tolist()
        exterior = row_col_list_to_points(exterior_np, do_round=True)
        return Polyline(exterior)

    @classmethod
    def allowed_transforms(cls):
        """
        """
        from supervisely.geometry.any_geometry import AnyGeometry
        from supervisely.geometry.rectangle import Rectangle
        from supervisely.geometry.alpha_mask import AlphaMask
        from supervisely.geometry.bitmap import Bitmap
        from supervisely.geometry.polygon import Polygon

        return [AnyGeometry, Rectangle, Bitmap, Polygon, AlphaMask]



from typing import List, Literal, Optional

from supervisely.app.widgets import Widget


class Container(Widget):
    """Container widget in Supervisely is a flexible tool that allows for organizing other widgets within it.

    Read about it in `Developer Portal <https://developer.supervisely.com/app-development/widgets/layouts-and-containers/container>`_
        (including screenshots and examples).

    :param widgets: list of widgets to be placed in the container
    :type widgets: Optional[List[Widget]]
    :param direction: direction of the container, one of: vertical, horizontal
    :type direction: Optional[Literal["vertical", "horizontal"]]
    :param gap: gap between widgets in pixels
    :type gap: Optional[int]
    :param fractions: list of fractions for each widget (only for horizontal direction)
    :type fractions: Optional[List[int]]
    :param overflow: overflow behavior, one of: scroll, wrap
    :type overflow: Optional[Literal["scroll", "wrap"]]
    :param style: CSS style for the container
    :type style: Optional[str]
    :param widget_id: An identifier of the widget.
    :type widget_id: str, optional

    :Usage example:
    .. code-block:: python

        from supervisely.app.widgets import Container, Text

        container = Container(
            widgets=[
                Text("First widget"),
                Text("Second widget"),
            ],
            direction="horizontal",
            gap=10,
            fractions=[1, 2],
            overflow="scroll",
            style="background-color: #f0f0f0; padding: 10px",
        )
    """

    def __init__(
        self,
        widgets: Optional[List[Widget]] = [],
        direction: Optional[Literal["vertical", "horizontal"]] = "vertical",
        gap: Optional[int] = 10,
        fractions: Optional[List[int]] = None,
        overflow: Optional[Literal["scroll", "wrap"]] = "scroll",
        style: Optional[str] = "",
        widget_id: Optional[str] = None,
        widgets_style: Optional[str] = "",
    ):
        self._widgets = widgets
        self._direction = direction
        self._gap = gap
        self._overflow = overflow
        self._style = style
        self._widgets_style = widgets_style

        if self._overflow not in ["scroll", "wrap", None]:
            raise ValueError("overflow can be only 'scroll', 'wrap' or None")

        if self._direction == "vertical" and self._overflow == "wrap":
            raise ValueError("overflow can be 'wrap' only with horizontal direction")

        if self._direction == "vertical" and fractions is not None:
            raise ValueError("fractions can be defined only with horizontal direction")

        if fractions is not None and len(widgets) != len(fractions):
            raise ValueError(
                "len(widgets) != len(fractions): fractions have to be defined for all widgets"
            )

        if self._direction == "vertical":
            self._overflow = None

        self._fractions = fractions
        self._flex_direction = "column"
        if direction == "horizontal":
            self._flex_direction = "row"
            if self._fractions is None:
                self._fractions = ["1 1 auto"] * len(self._widgets)
        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self) -> None:
        """The Container widget does not have any data
        the method is overridden to return None."""
        return None

    def get_json_state(self) -> None:
        """The Container widget does not have any state
        the method is overridden to return None."""
        return None



# coding: utf-8

from copy import deepcopy
from typing import List, Optional, Tuple, Union
from supervisely._utils import take_with_default, validate_img_size
from supervisely.video_annotation.frame_collection import FrameCollection
from supervisely.video_annotation.key_id_map import KeyIdMap
from supervisely.volume_annotation import constants
from supervisely.volume_annotation.volume_object_collection import VolumeObjectCollection
from supervisely.volume_annotation.slice import Slice
from supervisely.sly_logger import logger

# example:
# "volumeMeta": {
#     "ACS": "RAS",
#     "intensity": { "max": 3071, "min": -3024 },
#     "windowWidth": 6095,
#     "rescaleSlope": 1,
#     "windowCenter": 23.5,
#     "channelsCount": 1,
#     "dimensionsIJK": { "x": 512, "y": 512, "z": 139 },
#     "IJK2WorldMatrix": [
#         0.7617189884185793, 0, 0, -194.238403081894, 0, 0.7617189884185793, 0,
#         -217.5384061336518, 0, 0, 2.5, -347.7500000000001, 0, 0, 0, 1
#     ],
#     "rescaleIntercept": 0
# },


class Plane(FrameCollection):
    """
    A class representing a plane in medical image data.

    :param plane_name: Name of the plane, should be one of "sagittal", "coronal", "axial", or None for spatial figures.
    :type plane_name: Union[str, None]
    :param img_size: Size of the plane image
    :type img_size: Optional[Union[Tuple[int, int], None]]
    :param slices_count: Number of slices in the plane.
    :type slices_count: Optional[Union[int, None]]
    :param items: List of :py:class:`Slice<supervisely.volume_annotation.slice.Slice>` objects representing the slices in the plane.
    :type items: Oprional[List[:py:class:`Slice<supervisely.volume_annotation.slice.Slice>`]]
    :param volume_meta: Metadata of the volume.
    :type volume_meta: Optional[dict]
    :Usage example:

     .. code-block:: python

        import supervisely as sly
        from supervisely.volume_annotation.plane import Plane
        path = "/Users/almaz/Downloads/my volumes/ds11111/Demo volumes_ds1_CTChest.nrrd"
        vol, meta  = sly.volume.read_nrrd_serie_volume(path)

        plane = Plane(
            sly.Plane.AXIAL,
            volume_meta=meta,
        )
        print(plane.name) # axial
    """

    item_type = Slice

    SAGITTAL = "sagittal"
    """Sagittal plane of the volume."""

    CORONAL = "coronal"
    """Coronal plane of the volume."""

    AXIAL = "axial"
    """Axial plane of the volume."""

    _valid_names = [SAGITTAL, CORONAL, AXIAL, None]  # None for spatial figures

    @staticmethod
    def validate_name(name: Union[str, None]):
        """
        Validates if the given plane name is valid.

        :param name: Name of the plane.
        :type name: Union[str, None]

        :raises ValueError: If `name` is not one of "sagittal", "coronal", "axial", or None.
        :Usage example:

         .. code-block:: python

            import supervisely as sly
            from supervisely.volume_annotation.plane import Plane
            plane_name_1 = "axial"
            Plane.validate_name(plane_name_1)

            plane_name_2 = "xy"
            Plane.validate_name(plane_name_2)
            # ValueError: Unknown plane xy, valid names are ['sagittal', 'coronal', 'axial', None]
        """

        if name not in Plane._valid_names:
            raise ValueError(f"Unknown plane {name}, valid names are {Plane._valid_names}")

    def __init__(
        self,
        plane_name: Union[str, None],
        img_size: Union[tuple, list] = None,
        slices_count: int = None,
        items: Optional[List[Slice]] = None,
        volume_meta: Optional[dict] = None,
    ):
        Plane.validate_name(plane_name)
        self._name = plane_name

        if img_size is None and volume_meta is None:
            raise ValueError(
                "Both img_size and volume_meta are None, only one of them is allowed to be a None"
            )
        if slices_count is None and volume_meta is None:
            raise ValueError(
                "Both slices_count and volume_meta are None, only one of them is allowed to be a None"
            )
        self._img_size = take_with_default(img_size, Plane.get_img_size(self._name, volume_meta))
        self._img_size = validate_img_size(self._img_size)

        self._slices_count = take_with_default(
            slices_count, Plane.get_slices_count(self._name, volume_meta)
        )

        super().__init__(items=items)

    @property
    def name(self) -> Union[str, None]:
        """
        Get the name of the plane.

        :return: Name of the plane.
        :rtype: Union[str, None]
        :Usage example:

         .. code-block:: python

            import supervisely as sly
            path = "/Users/almaz/Downloads/my volumes/ds11111/Demo volumes_ds1_CTChest.nrrd"
            vol, meta = sly.volume.read_nrrd_serie_volume(path)

            plane = sly.Plane(
                sly.Plane.AXIAL,
                volume_meta=meta,
            )
            print(plane.name)
            # Output: axial
        """

        return self._name

    @property
    def slices_count(self) -> int:
        """
        Get the number of slices in the plane.

        :return: Number of slices in the plane.
        :rtype: int
        :Usage example:

         .. code-block:: python

            import supervisely as sly
            path = "/Users/almaz/Downloads/my volumes/ds11111/Demo volumes_ds1_CTChest.nrrd"
            vol, meta = sly.volume.read_nrrd_serie_volume(path)

            plane = sly.Plane(
                sly.Plane.AXIAL,
                volume_meta=meta,
            )
            print(plane.slices_count)
            # Output: 139
        """

        return self._slices_count

    @property
    def img_size(self) -> Tuple[int]:
        """
        Get the size of the image in the plane.

        :return: Size of the image in the plane.
        :rtype: Tuple[int]
        :Usage example:

         .. code-block:: python

            import supervisely as sly
            path = "/Users/almaz/Downloads/my volumes/ds11111/Demo volumes_ds1_CTChest.nrrd"
            vol, meta = sly.volume.read_nrrd_serie_volume(path)

            plane = sly.Plane(
                sly.Plane.AXIAL,
                volume_meta=meta,
            )
            print(plane.img_size)
            # Output: (512, 512)
        """

        return deepcopy(self._img_size)

    @property
    def normal(self) -> dict:
        """
        Returns the normal vector of the plane.

        :return: A dictionary representing the normal vector of the plane.
        :rtype: dict
        :Usage example:

         .. code-block:: python

            import supervisely as sly
            path = "/Users/almaz/Downloads/my volumes/ds11111/Demo volumes_ds1_CTChest.nrrd"
            vol, meta = sly.volume.read_nrrd_serie_volume(path)

            plane = sly.Plane(
                sly.Plane.AXIAL,
                volume_meta=meta,
            )
            print(plane.normal)
            # Output: {'x': 0, 'y': 0, 'z': 1}
        """

        return Plane.get_normal(self.name)

    def __str__(self):
        return super().__str__().replace("Frames", "Slices")

    @staticmethod
    def get_normal(name: str) -> dict:
        """
        Returns the normal vector of a plane given its name.

        :param name: Name of the plane.
        :type name: str
        :return: A dictionary representing the normal vector of the plane.
        :rtype: dict
        :raises ValueError: If `name` is not one of "sagittal", "coronal", or "axial".
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            print(sly.Plane.get_normal(sly.Plane.AXIAL))
            # Output: {'x': 0, 'y': 0, 'z': 1}
        """

        Plane.validate_name(name)
        if name == Plane.SAGITTAL:
            return {"x": 1, "y": 0, "z": 0}
        if name == Plane.CORONAL:
            return {"x": 0, "y": 1, "z": 0}
        if name == Plane.AXIAL:
            return {"x": 0, "y": 0, "z": 1}

    @staticmethod
    def get_name(normal: dict) -> str:
        """
        Returns the name of a plane given its normal vector.

        :param normal: A dictionary representing the normal vector of a plane.
        :type normal: dict
        :return: The name of the plane.
        :rtype: str
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            print(sly.Plane.get_name({'x': 0, 'y': 0, 'z': 1}))
            # Output: axial
        """

        if normal == {"x": 1, "y": 0, "z": 0}:
            return Plane.SAGITTAL
        if normal == {"x": 0, "y": 1, "z": 0}:
            return Plane.CORONAL
        if normal == {"x": 0, "y": 0, "z": 1}:
            return Plane.AXIAL

    @staticmethod
    def get_img_size(name: str, volume_meta: dict) -> List[int]:
        """
        Get size of the image for a given plane and volume metadata.

        :param name: Name of the plane.
        :type name: str
        :param volume_meta: Metadata for the volume.
        :type volume_meta: dict
        :return: The size of the image for the given plane.
        :rtype: List[int]
        :raises ValueError: If `name` is not one of "sagittal", "coronal", or "axial".
        :Usage example:

         .. code-block:: python

            import supervisely as sly
            path = "/Users/almaz/Downloads/my volumes/ds11111/Demo volumes_ds1_CTChest.nrrd"
            vol, meta = sly.volume.read_nrrd_serie_volume(path)

            plane = sly.Plane(
                sly.Plane.AXIAL,
                volume_meta=meta,
            )
            img_size = sly.Plane.get_img_size(plane.name, meta)
            print(img_size)
            # Output: [512, 512]
        """

        Plane.validate_name(name)
        dimentions = volume_meta["dimensionsIJK"]
        # (height, width)
        height = None
        width = None
        if name == Plane.SAGITTAL:
            width = dimentions["y"]
            height = dimentions["z"]
        elif name == Plane.CORONAL:
            width = dimentions["x"]
            height = dimentions["z"]
        elif name == Plane.AXIAL:
            width = dimentions["x"]
            height = dimentions["y"]
        return [height, width]

    @staticmethod
    def get_slices_count(name: str, volume_meta: dict) -> int:
        """
        Returns the number of slices in the given plane.

        :param name: Name of the plane.
        :type name: str
        :param volume_meta: Metadata of the volume to extract slices from.
        :type volume_meta: dict
        :return: Number of slices in the given plane.
        :rtype: int
        :raises ValueError: If `name` is not one of "sagittal", "coronal", or "axial".
        :Usage example:

         .. code-block:: python

            import supervisely as sly
            path = "/Users/almaz/Downloads/my volumes/ds11111/Demo volumes_ds1_CTChest.nrrd"
            vol, meta = sly.volume.read_nrrd_serie_volume(path)

            plane = sly.Plane(
                sly.Plane.AXIAL,
                volume_meta=meta,
            )
            slices_count = sly.Plane.get_slices_count(plane.name, meta)
            print(slices_count)
            # Output: 139
        """

        Plane.validate_name(name)
        dimentions = volume_meta["dimensionsIJK"]
        if name == Plane.SAGITTAL:
            return dimentions["x"]
        elif name == Plane.CORONAL:
            return dimentions["y"]
        elif name == Plane.AXIAL:
            return dimentions["z"]

    @classmethod
    def from_json(
        cls,
        data: dict,
        plane_name: str,
        objects: VolumeObjectCollection,
        img_size: Optional[Union[list, tuple]] = None,
        slices_count: Optional[int] = None,
        volume_meta: Optional[dict] = None,
        key_id_map: Optional[KeyIdMap] = None,
    ):
        """
        Creates a `Plane` instance from a JSON dictionary.

        :param data: JSON dictionary representing a `Plane` instance.
        :type data: dict
        :param plane_name: Name of the plane.
        :type plane_name: str
        :param objects: Objects in the plane.
        :type objects: VolumeObjectCollection
        :param img_size: Size of the image represented by the plane.
        :type img_size: Optional[Union[list, tuple]]
        :param slices_count: Number of slices along the plane.
        :type slices_count: Optional[int]
        :param volume_meta: Metadata of the volume to extract slices from.
        :type volume_meta: Optional[dict]
        :param key_id_map: Dictionary mapping object keys to object IDs.
        :type key_id_map: Optional[KeyIdMap]
        :return: A new class:`Plane<Plane>` instance created from the JSON.
        :rtype: :py:class:`Plane<supervisely.volume_annotation.plane.Plane>`

        :raises ValueError: If `plane_name` is not equal to the "name" field in "data", or if the "normal" field in "data" is not valid for the given plane, or if both `slices_count` and `volume_meta` are None.
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            # Create Plane from json we use data from example to_json(see below)
            path = "/Users/almaz/Downloads/my volumes/ds11111/Demo volumes_ds1_CTChest.nrrd"
            new_plane = sly.Plane.from_json(
                data=plane.to_json(),
                plane_name=sly.Plane.AXIAL,
                objects=sly.VolumeObjectCollection([]),
                volume_meta=meta,
            )
        """

        Plane.validate_name(plane_name)
        if plane_name != data[constants.NAME]:
            raise ValueError(
                f"Plane name {plane_name} differs from the name in json data {data[constants.NAME]}"
            )

        if Plane.get_normal(plane_name) != data[constants.NORMAL]:
            raise ValueError(
                f"Normal json data {data[constants.NORMAL]} is not valid for {plane_name}. It should be {Plane.get_normal(plane_name)}"
            )
        if slices_count is None:
            if volume_meta is None:
                raise ValueError(
                    "Both slices_count and volume_meta are None, only one of them is allowed to be a None"
                )
            else:
                slices_count = Plane.get_slices_count(plane_name, volume_meta)

        slices = []
        for slice_json in data[constants.SLICES]:
            slices.append(
                cls.item_type.from_json(slice_json, objects, plane_name, slices_count, key_id_map)
            )

        return cls(plane_name, img_size, slices_count, slices, volume_meta)

    def to_json(
        self,
        key_id_map: Optional[KeyIdMap] = None,
    ) -> dict:
        """
        Returns a JSON serializable dictionary representation of the `Plane` instance.

        :param key_id_map: Dictionary mapping object keys to object IDs.
        :type key_id_map: Optional[KeyIdMap]
        :return: A JSON serializable dictionary representation of the `Plane` instance.
        :rtype: dict
        :Usage example:

         .. code-block:: python

            import supervisely as sly
            from supervisely.volume_annotation.plane import Plane
            path = "/Users/almaz/Downloads/my volumes/ds11111/Demo volumes_ds1_CTChest.nrrd"
            vol, meta = sly.volume.read_nrrd_serie_volume(path)

            plane = Plane(
                sly.Plane.AXIAL,
                volume_meta=meta,
            )
            print(plane.to_json())
            # Output: { "name": "axial", "normal": { "x": 0, "y": 0, "z": 1 }, "slices": [] }
        """
        json_slices = []
        for slice in self:
            slice: Slice
            json_slices.append(slice.to_json(key_id_map))

        return {
            constants.NAME: self.name,
            constants.NORMAL: self.normal,
            constants.SLICES: json_slices,
        }

    def validate_figures_bounds(self):
        """
        Validates the figure bounds for all slices in the Plane.

        :raises ValueError: If any of the slices in the `Plane` instance have invalid figure bounds.
        """
        for slice in self:
            slice: Slice
            slice.validate_figures_bounds(self.img_size)



from supervisely.sly_logger import logger

try:
    from supervisely.nn.tracker.bot_sort import BoTTracker
    from supervisely.nn.tracker.deep_sort import DeepSortTracker
except ImportError:
    logger.error(
        "Failed to import tracker modules. Please try install extras with 'pip install supervisely[tracking]'"
    )
    raise



from supervisely import logger

import numpy as np


def las2pcd(input_path, output_path):
    import laspy # pylint: disable=import-error
    import open3d as o3d # pylint: disable=import-error

    try:
        las = laspy.read(input_path)
    except Exception as e:
        if "buffer size must be a multiple of element size" in str(e):
            from laspy.point.record import PackedPointRecord # pylint: disable=import-error
            logger.warn(
                "Could not read LAS file in laspy. Trying to read it without EXTRA_BYTES..."
            )

            @classmethod
            def from_buffer_without_extra_bytes(cls, buffer, point_format, count=-1, offset=0):
                item_size = point_format.size
                count = len(buffer) // item_size
                points_dtype = point_format.dtype()
                data = np.frombuffer(buffer, dtype=points_dtype, offset=offset, count=count)
                return cls(data, point_format)
            PackedPointRecord.from_buffer = from_buffer_without_extra_bytes
            las = laspy.read(input_path)
        else:
            raise e
    point_cloud = np.vstack((las.X, las.Y, las.Z)).T
    pc = o3d.geometry.PointCloud(o3d.utility.Vector3dVector(point_cloud))
    o3d.io.write_point_cloud(output_path, pc)


# coding: utf-8
from supervisely.annotation.obj_class import ObjClass
from supervisely.annotation.obj_class_collection import ObjClassCollection
from supervisely.annotation.renamer import Renamer


class ObjClassMapper:
    """
    """
    def map(self, src: ObjClass) -> ObjClass:
        """
        """
        raise NotImplementedError()


class RenamingObjClassMapper(ObjClassMapper):
    """
    This is a class for renaming ObjClass in given ObjClassCollection
    """
    def __init__(self, dest_obj_classes: ObjClassCollection, renamer: Renamer):
        self._dest_obj_classes = dest_obj_classes
        self._renamer = renamer

    def map(self, src: ObjClass) -> ObjClass:
        """
        The function map rename ObjClass in given collection
        :return: ObjClass
        """
        dest_name = self._renamer.rename(src.name)
        return self._dest_obj_classes.get(dest_name, None) if (dest_name is not None) else None



from typing import Optional, Union, List
from supervisely.app.widgets import Widget, NotificationBox, Button, generate_id
from supervisely import ObjClass, ObjClassCollection
from supervisely.app import DataJson, StateJson

from supervisely.geometry.bitmap import Bitmap
from supervisely.geometry.cuboid import Cuboid
from supervisely.geometry.point import Point
from supervisely.geometry.polygon import Polygon
from supervisely.geometry.polyline import Polyline
from supervisely.geometry.rectangle import Rectangle
from supervisely.geometry.graph import GraphNodes
from supervisely.geometry.any_geometry import AnyGeometry
from supervisely.geometry.cuboid_3d import Cuboid3d
from supervisely.geometry.pointcloud import Pointcloud
from supervisely.geometry.point_3d import Point3d
from supervisely.geometry.multichannel_bitmap import MultichannelBitmap
from supervisely.geometry.closed_surface_mesh import ClosedSurfaceMesh


type_to_shape_text = {
    AnyGeometry: "any shape",
    Rectangle: "rectangle",
    Polygon: "polygon",
    Bitmap: "bitmap (mask)",
    Polyline: "polyline",
    Point: "point",
    Cuboid: "cuboid",  #
    Cuboid3d: "cuboid 3d",
    Pointcloud: "pointcloud",  #  # "zmdi zmdi-border-clear"
    MultichannelBitmap: "n-channel mask",  # "zmdi zmdi-collection-item"
    Point3d: "point 3d",  # "zmdi zmdi-select-all"
    GraphNodes: "keypoints",
    ClosedSurfaceMesh: "volume (3d mask)",
}


class ClassesMapping(Widget):
    def __init__(
        self,
        classes: Optional[Union[List[ObjClass], ObjClassCollection]] = [],
        empty_notification: Optional[NotificationBox] = None,
        widget_id: Optional[str] = None,
    ):
        if empty_notification is None:
            empty_notification = NotificationBox(
                title="No classes",
                description="No classes to map.",
            )
        self.empty_notification = empty_notification
        self._classes = classes

        self._select_all_btn = Button()
        self._deselect_all_btn = Button()

        self._select_all_btn = Button(
            "Select all",
            button_type="text",
            show_loading=False,
            icon="zmdi zmdi-check-all",
            widget_id=generate_id(),
        )
        self._deselect_all_btn = Button(
            "Deselect all",
            button_type="text",
            show_loading=False,
            icon="zmdi zmdi-square-o",
            widget_id=generate_id(),
        )

        @self._select_all_btn.click
        def _select_all_btn_clicked():
            self.select_all()

        @self._deselect_all_btn.click
        def _deselect_all_btn_clicked():
            self.deselect_all()

        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self):
        return {
            "classes": [
                {
                    **cls.to_json(),
                    "shape_text": type_to_shape_text.get(cls.geometry_type).upper(),
                    "default_value": cls.name,
                }
                for cls in self._classes
            ]
        }

    def get_json_state(self):
        return {
            "classes_values": [
                {
                    "value": cls.name,
                    "default": True,
                    "ignore": False,
                    "selected": False,
                }
                for cls in self._classes
            ]
        }

    def set(self, classes):
        self._classes = classes
        self.update_data()
        DataJson().send_changes()
        cur_mapping = self.get_mapping()
        new_mapping_values = []
        for cls in self._classes:
            value = cur_mapping.get(
                cls.name,
                {
                    "value": cls.name,
                    "default": False,
                    "ignore": True,
                    "selected": False,
                },
            )
            new_mapping_values.append(value)
        StateJson()[self.widget_id]["classes_values"] = new_mapping_values
        StateJson().send_changes()

    def get_classes(self):
        return self._classes

    def get_mapping(self):
        classes_values = StateJson()[self.widget_id]["classes_values"]
        if len(classes_values) != len(self._classes):
            self.update_state()
            return self.get_mapping()
        mapping = {
            cls.name: classes_values[idx]
            for idx, cls in enumerate(self._classes)
            if classes_values[idx]["selected"]
        }
        return mapping

    def ignore(self, indexes: List[int]):
        classes_values = StateJson()[self.widget_id]["classes_values"]
        for idx in indexes:
            classes_values[idx] = {"value": "", "default": False, "ignore": True}
        StateJson()[self.widget_id]["classes_values"] = classes_values
        StateJson().send_changes()

    def set_default(self):
        self.update_state()
        StateJson().send_changes()

    def set_mapping(self, mapping: dict):
        cur_mapping = self.get_mapping()
        new_mapping_values = []
        for cls in self._classes:
            cur_value = cur_mapping.get(cls.name, {"value": ""}).get("value")
            new_value = mapping.get(cls.name, cur_value)
            new_mapping_values.append(
                {
                    "value": new_value if new_value != "" else cls.name,
                    "default": new_value == cls.name,
                    "ignore": new_value == "",
                    "selected": new_value != "",
                }
            )
        StateJson()[self.widget_id]["classes_values"] = new_mapping_values
        StateJson().send_changes()

    def select_all(self):
        classes_values = StateJson()[self.widget_id]["classes_values"]
        classes_values: dict
        for value in classes_values:
            value["selected"] = True
        StateJson()[self.widget_id]["classes_values"] = classes_values
        StateJson().send_changes()

    def deselect_all(self):
        classes_values = StateJson()[self.widget_id]["classes_values"]
        classes_values: dict
        for value in classes_values:
            value["selected"] = False
        StateJson()[self.widget_id]["classes_values"] = classes_values
        StateJson().send_changes()

    def select(self, classes):
        classes_values = StateJson()[self.widget_id]["classes_values"]
        classes_values: list
        for idx, cls in enumerate(self._classes):
            classes_values[idx]["selected"] = cls.name in classes
        StateJson()[self.widget_id]["classes_values"] = classes_values
        StateJson().send_changes()
